quiet = False
log = '7;37;40'
info = '7;32;40'
warn = '7;33;40'
error = '7;31;40'
state = '7;34;47'

class Settings:
    quiet = False


def mprint(text: str, type: str):
    if not Settings.quiet:
        print(f'\x1b[{type}m {text} \x1b[0m')

def ctext(text : str, type : str)->str:
    return f'\x1b[{type}m {text} \x1b[0m'



#!/usr/bin/env python
import argparse
from smurfs._smurfs.smurfs import Smurfs
from smurfs._smurfs.multi_smurfs import MultiSmurfs
from smurfs.__version__ import __version__
from smurfs.support.mprint import *
import os
import pickle
import subprocess
import sys
import warnings


def get_parser():
    parser = argparse.ArgumentParser()

    parser.add_argument("target", help="This parameter represents the target analyzed by SMURFS. This parameter is "
                                       "rather vague, to give you the maximum amount of flexibility. Either you can "
                                       "provide any name of a star (resolvable through Simbad) that has been observed "
                                       "by TESS/Kepler/K2. You can also provide a filename through this parameter.\n\n"
                                       "If you choose to do the the first, the code prioritizes TESS over all "
                                       "other missions (if you don't provide the mission parameter), and SC over LC "
                                       "observations. It then checks MAST if there are SC observations "
                                       "from TESS for this star. If there are, it uses "
                                       "*lightkurve.search_lightcurvefile* to download those."
                                       " If there are none, it checks if there are "
                                       "LC observations from TESS and uses Eleanor to extract the light curves "
                                       "from the FFIs. If there are None, it searches all other missions "
                                       "for light kurves of this object.If you choose another "
                                       " mission, it will call *lightkurve.search_lightcurvefile* for this specific "
                                       "mission.\n\nIf you choose to provide "
                                       "a file through this parameter, make sure that you follow these conventions:\n"
                                       "- The file must be an ASCII file\n- The first column must contain the "
                                       "time stamps\n- The second column must contain the flux\n* If a third "
                                       "column exists, SMURFS will assume that these are the uncertainties "
                                       "in the flux.\n\nSMURFS will take the file as is and won't apply "
                                       "any corrections to it, if you don't use the *apply_corrections* flag. "
                                       "It then assumes that the flux values are in magnitude and varying "
                                       "around 0, that the time stamps are in days and that the "
                                       "data set is properly reduced.", type=str)
    parser.add_argument("snr",
                        help="This parameter represents the lower bound signal to noise ratio any frequency must have. SMURFS"
                             "computes the SNR by taking the amplitude of an individual frequency (as defined by the amplitude in the frequency"
                             "spectrum). Next, it applies half of the window size to either side of the frequency, starting by the next adjacent"
                             "minima, and takes the mean of this window as the noise surrounding the frequency. The ratio of these two values is"
                             "the resulting SNR for an individual frequency. By default, SMURFS stops its run when the first frequency less than"
                             "this value has been found",
                        type=float)
    parser.add_argument("windowSize", help="The window size used to get the SNR for a given frequency. SMURFS "
                                           "defines the window as half of the given value on each side of the peak "
                                           "in the periodogram, starting from the first minima next to the peak.",
                        type=float)
    parser.add_argument("-fr", "--frequencyRange",
                        help="Setting this parameter, allows you to restrict the analysis"
                             "of a time series to a given range in the frequency spectrum. This might be useful, if yo have a lot of high"
                             "amplitude noise in another part of the spectrum, than the one you might be interested in. Provide the parameter "
                             "in the form *0,100*.",
                        type=str, default="None,None")

    parser.add_argument("-ssa", "--skipSimilarFrequencies",
                        help="Ignores regions where SMURFS finds multiple frequencies within a small range."
                             "This can happen due to insufficient fits or signals that are hard to fit.",
                        action='store_true')

    parser.add_argument("-sc", "--skipCutoff",
                        help="Skips the cutoff check, so similar frequencies don't chancel the run. Be aware that this "
                             "might lead to unknown behaviour",
                        action='store_true')

    parser.add_argument("-ef", "--extendFrequencies",
                        help="Extends the analysis by n frequencies. By default, SMURFS stops when it finds the first "
                             "insignificant frequency. Setting this parameter requires SMURFS to find n insignificant "
                             "frequencies in a row.", type=int, default=0)

    parser.add_argument("-imf", "--improveFitMode",
                        help="This parameter defines the way SMURFS uses a Period like improvement of frequencies. You "
                             "can set the following modes:\n* -*all*: Tries to refit all found frequencies after every "
                             "frequeny that is found. This is the usual behaviour of Period and the default setting.\n"
                             "-*end*: Improves the frequencies after SMURFS would stop its run\n-*none*: Disables "
                             "the improve frequencies setting. This can be useful if you find a lot of frequencies and "
                             "the run would take an unnecessary amount of time",type=str, choices=['all', 'end', 'none'],
                        default='all')

    parser.add_argument("-fm", "--fitMethod",
                        help="SMURFS implements two different fitting libraries: Either it uses *scipy* "
                             "(scipy.optimize.curve_fit) or *lmfit* (lmfit Model fit). Choosing one over the other "
                             "might lead to different results, so if you have find an unexpected result, try "
                             "to switch the fitting mode.",
                        type=str, choices=["scipy", "lmfit"], default="lmfit")

    parser.add_argument("-ft", "--fluxType",
                        help='Due to how TESS reduces its data, you can choose different flux types if you provide '
                             'a star name. For SC data, TESS gives you two possibilities:\n- *SAP flux*: SAP is the '
                             'simple aperture photometry flux (resulting light curve is the flux after summing the '
                             'calibrated pixels within the TESS optimal aperture)\n- *PDCSAP flux*: PDCSAP is the '
                             'Pre-search Data conditioned Simple aperture photometry, which is corrected using '
                             'co-trending basis vectors.\n\nBy default we use the PDCSAP flux, but you can also '
                             'choose another one if you like.\n\nIf your target is only observed in LC mode, SMURFS '
                             'also provides these two modes (these have a slightly different meaning, their '
                             'result is however equivalent to the SC SAP and PDSCAP fluxes, see the Eleanor '
                             'documentation). However, in LC mode you have also the PSF flux type, which models a '
                             'point spread function for a given star. The validation page generated when SMURFS '
                             'extracts a target from a FFI always shows the SAP and the chosen flux to compare. You '
                             'might want to try different settings, depending on your use case',
                        type=str, choices=["PDCSAP", "SAP", "PSF"], default="PDCSAP")
    parser.add_argument("-pca", "--do_pca",
                        help="Activates the PCA analysis (aperture × TPF + background subtraction + "
                             "cotrending basis vectors). This doesn't change the data you are trying to analyze, "
                             "but shows the PDSCAP flux in the validation page. Only applicable "
                             "when using LC data.", action='store_true')
    parser.add_argument("-psf", "--do_psf", help="Activates the PSF analysis. This adds point spread function "
                                                 "modelling to the extraction of light curves from FFIs.  This doesn't "
                                                 "change the data you are trying to analyze, "
                                                 "but shows the PSF flux in the validation page. Only applicable "
                                                 "when using LC data.", action='store_true')

    parser.add_argument("-so", "--storeObject", help="If this flag is set, the SMURFS object is stored in the "
                                                     "results. You can later use this object to load the "
                                                     "_result into a python file using 'Smurfs.from_path'. Allows "
                                                     "you to easily access all convenience functions from SMURFS."
                        , action='store_true')

    parser.add_argument("-sp", "--savePath", help="Allows you to set the save path of the analysis. By "
                                                  "default it will save it in the same folder, where "
                                                  "the module was called.", type=str, default=".")

    parser.add_argument("-i", "--interactive", help="Using this flag will automatically start an iPython shell. "
                                                    "This allows for direct interaction with the result using the "
                                                    "'star' object. You can then access all convenience functions "
                                                    "directly (like plotting, the FrequencyFinder object, etc)."
                        , action='store_true')

    parser.add_argument("-m", "--mission", help="Three different missions are available: Kepler,TESS,K2. You can choose"
                                                " the mission by setting this value. By default, only TESS missions are"
                                                " considered"
                        , type=str, choices=["Kepler", "TESS", "K2"], default="TESS")

    parser.add_argument("-cl", "--sigmaClip", help="Sets the sigma for the sigma clipping. Default is 4.", type=float,
                        default=4)
    parser.add_argument("-it", "--iters", help="Sets the iterations for the sigma clipping. Default is 1.", type=int,
                        default=1)

    parser.add_argument("-ac", "--apply_corrections",
                        help="If this flag is set, correction (sigma clipping, conversion"
                             " to magnitude) are applied to files. Make sure that your flux is in "
                             "electron counts if you use this flag.", action='store_true')

    """
    #todo replace this
    parser.add_argument("-om", "--outputMode", help="Optional parameter that defines what will be exported as dataset. "
                                                    "There are two possibilities: Normal and Full. Normal mode means, that"
                                                    "only the first and last amplitudespectra will be plotted & exported, "
                                                    "full means all amplitudespectra will be plotted & exported. Default is "
                                                    "Normal.",
    
                        type=str, choices=["Normal", "Full"], default="Normal")
                        
    parser.add_argument("-igr", "--ignoreCutoffRatio", help="Optional parameter. If this is set to True, it will ignore"
                                                            "the gap ratio cutoff criterion", action='store_true')  
                                                            
    parser.add_argument("-trs", "--timeBaseSplit",
                        help="Optional parameter that describes the split within the time range."
                             "This can be used to get multiple amplitude spectra for the "
                             "lightcurve, split up within the time Range. A parameter could be "
                             "for example 50, where the parameter would split the lightcurve in"
                             "50 day chunks (last segment will be smaller/equal). Defaults"
                             "to -1, which means no splitting will occur",
                        type=int, default=-1)
    parser.add_argument("-o", "--overlap",
                        help="Optional parameter that describes the overlap of two split ranges. Will only"
                             "work, if -trs is set to a reasonable parameter. For example, if the trs is "
                             "set to 50 and the overlap to 2, the chunks would contain the ranges (0,50),"
                             "(48,98),(96,146) ... Defaults to 0, which means no overlap will take place",
                        type=int, default=0)                                                                                  
    """
    parser.add_argument("--version", help="Shows version of SMURFS", action='version',
                        version='SMURFS {version}'.format(version=__version__))
    return parser


def main(args=None):
    if args is None:
        args = sys.argv[1:]

    parser = get_parser()
    args = parser.parse_args(args)

    if "," in args.target:
        targets = args.target.split(",")
    else:
        targets = [args.target]

    f_min = None if args.frequencyRange.split(",")[0] == 'None' else float(args.frequencyRange.split(",")[0])
    f_max = None if args.frequencyRange.split(",")[1] == 'None' else float(args.frequencyRange.split(",")[1])

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        if len(targets) == 1:
            target = targets[0]

            if os.path.exists(target) and os.path.isfile(target):
                s = Smurfs(file=target, flux_type=args.fluxType, mission=args.mission, sigma_clip=args.sigmaClip,
                           iters=args.iters, do_pca=args.do_pca, do_psf=args.do_psf,
                           apply_file_correction=args.apply_corrections)
            else:
                s = Smurfs(target_name=target, flux_type=args.fluxType, mission=args.mission, sigma_clip=args.sigmaClip,
                           iters=args.iters, do_pca=args.do_pca, do_psf=args.do_psf)

            improve_fit = args.improveFitMode == 'all'

            s.run(snr=args.snr, window_size=args.windowSize, f_min=f_min, f_max=f_max,
                  skip_similar=args.skipSimilarFrequencies, similar_chancel=not args.skipCutoff
                  , extend_frequencies=args.extendFrequencies, improve_fit=improve_fit
                  , mode=args.fitMethod)

            if args.improveFitMode == 'end':
                mprint("Improving fit ...", state)
                s.improve_result()

            s.save(args.savePath, args.storeObject)

            # start interactive shell, by loading smurfs object directly into the shell
            if args.interactive:
                mprint("Starting interactive shell. Use the 'star' object to interact with the result!", state)
                pickle.dump(s, open("i_obj.smurfs", "wb"))
                cmd = ["ipython", "-i", "-c",
                       "from smurfs import Smurfs;import pickle;star : Smurfs = pickle.load(open('i_obj.smurfs', 'rb'));import os;os.remove('i_obj.smurfs');import matplotlib.pyplot as pl;pl.ion()"]
                subprocess.call(cmd)
                mprint("Done", info)
        else:
            if len(targets[0].split(".")) == 2 and os.path.basename(targets[0]).split(".")[1] in ['txt', 'dat']:
                s = MultiSmurfs(file_list=targets)
            else:
                s = MultiSmurfs(target_list=targets, flux_types=args.fluxType)

            s.run(snr=args.snr, window_size=args.windowSize, f_min=f_min, f_max=f_max,
                  skip_similar=args.skipSimilarFrequencies, similar_chancel=not args.skipCutoff
                  , extend_frequencies=args.extendFrequencies, improve_fit=args.disableImproveFrequencies
                  , mode=args.fitMethod)
            s.save(args.savePath, args.storeObject)


if __name__ == "__main__":
    main()



import numpy as np
from lightkurve import Periodogram as lk_Periodogram,LightCurve
from astropy.stats import LombScargle
from astropy.units import cds
from pandas import DataFrame as df
from typing import List,Tuple


class Periodogram(lk_Periodogram):
    """
    Custom Periodogram class, fit to the needs of smurfs. Mirrors the behaviour of the Lightkurve Periodogram class,
    and is derived from it. See
    https://docs.lightkurve.org/api/lightkurve.periodogram.Periodogram.html#lightkurve.periodogram.Periodogram for
    documentation on the constructor parameters.

    This class differs from the Lightkurve Periodogram class through two aspects:
    1) It adds a different static method, that converts a Lightcurve object into a periodogram.
    2) The plotting and saving of data has been adapted to fit the needs of smurfs
    """
    def __init__(self, frequency: np.ndarray, power: np.ndarray, nyquist: float = None, targetid=None,
                 default_view='frequency', meta={}):
        self.nyquist = nyquist
        super().__init__(frequency,power,nyquist=nyquist,targetid=targetid,default_view=default_view,meta=meta)

    @staticmethod
    def from_lightcurve(lc : LightCurve, f_min=None, f_max=None, remove_ranges : List[Tuple[float]] = None
                        , samples_per_peak = 10):

        """
        Computes a periodogram from a Lightcurve object and normalizes it according to Parcivals theorem. It then
        reflects the physical values in the Light curve and has the same units. It then returns a Periodogram object.

        It also has a possibility to remove certain ranges from the periodogram.
        :param lc: Lightcurve object
        :param f_min: Lower range for the periodogram
        :param f_max: Upper range for the periodogram
        :param remove_ranges: List of tuples, that represent areas in the periodogram that are ignored. These are
        removed from the periodogram
        :param samples_per_peak: number of samples per peak
        :return: Periodogram object
        """

        nyquist = 1/(2*np.median(np.diff(lc.time)))

        try:
            time = lc.time.value
        except AttributeError:
            time = lc.time

        try:
            flux = lc.flux.value
        except AttributeError:
            flux = lc.flux


        ls = LombScargle(time,flux,normalization='psd')

        if f_max is not None and f_max > nyquist:
            #todo warning here!
            pass

        if f_min is None:
            f_min = 0

        if f_max is None:
            f_max = nyquist

        f, p = ls.autopower(minimum_frequency=f_min, maximum_frequency=f_max,
                            samples_per_peak=samples_per_peak, nyquist_factor=1)

        # normalization of psd in order to get good amplitudes
        p = np.sqrt(4 / len(time)) * np.sqrt(p)

        # removing first item
        p = p[1:]
        f = f[1:]

        if remove_ranges is not None:
            mask_list = []
            for r in remove_ranges:
                mask = f < r[0]
                mask = np.logical_or(mask,f > r[1])
                mask_list.append(mask)

            combined_mask = mask_list[0]
            if len(mask_list) > 1:
                for m in mask_list[1:]:
                    combined_mask = np.logical_and(combined_mask,m)

            f = f[combined_mask]
            p = p[combined_mask]

        return Periodogram(f*(1/cds.d),p*cds.ppm,nyquist=nyquist,targetid=lc.targetid)

    def plot(self,scale='linear', ax=None, xlabel=None, ylabel=None, title='', style='lightkurve', view=None,
             unit=None,color='k', **kwargs):
        """
        Plots the periodogram. Same call signature as lightkurve.periodogram.Periodogram.
        """

        if 'linestyle' in kwargs.keys():
            ls = kwargs['linestyle']
            del kwargs['linestyle']
        else:
            ls = '-'

        if ylabel is not None:
            ylabel = ylabel
        elif 'ylabel' in kwargs.keys():
            ylabel = kwargs['ylabel']
            del kwargs['ylabel']
        else:
            ylabel='Amplitude [mag]'


        return super().plot(scale,ax,xlabel,ylabel,title,style,view,unit,linestyle=ls,color=color,**kwargs)

    def to_csv(self,file):
        """
        Stores the periodogram into a file.
        :param file: File object
        """
        frame = df.from_dict({'Frequency':self.frequency,'Power':self.power})
        frame.to_csv(file,index=False)


import pytest
from lightkurve import search_lightcurvefile
from smurfs._smurfs.frequency_finder import sin_multiple,Frequency
from smurfs._smurfs.smurfs import Smurfs
from smurfs.preprocess.tess import download_lc
import numpy as np

param_list = [
    ([0.6,0.2,0],(0.6,0.2,0))
    ,([0.6,0.5,0,0.3,3,0.1],(0.6,0.5,0))
    #,([0.6,0.2,0,0.1,6,0,10,5,6],(10,5,6))
]

@pytest.mark.skip(reason="Temporary Skip")
@pytest.mark.parametrize('value',param_list)
def testFrequency(value):
    x = np.linspace(0,100,num=5000)
    noise = np.random.normal(0,0.05,len(x))
    y = sin_multiple(x, *value[0]) + noise

    f = Frequency(x,y,4,2)
    f.pre_whiten(mode='scipy')

    assert np.abs(f.amp.nominal_value - value[1][0]) < 4 * f.amp.std_dev
    assert np.abs(f.f.nominal_value - value[1][1]) < 4 * f.f.std_dev
    assert np.abs(f.phase.nominal_value - value[1][2]) < 8 * f.phase.std_dev

    f = Frequency(x, y, 4, 2)
    f.pre_whiten(mode='lmfit')

    assert np.abs(f.amp.nominal_value - value[1][0]) < 8 * f.amp.std_dev
    assert np.abs(f.f.nominal_value - value[1][1]) < 8 * f.f.std_dev
    try:
        assert np.abs(f.phase.nominal_value - value[1][2]) < 8 * f.phase.std_dev
    except AssertionError:
        assert np.abs(f.phase.nominal_value - (value[1][2] + 2 * np.pi)) < 8 * f.phase.std_dev



import pytest
from lightkurve import search_lightcurvefile
from smurfs._smurfs.frequency_finder import FFinder,sin,sin_multiple
from smurfs._smurfs.smurfs import Smurfs
from smurfs.preprocess.tess import download_lc
import numpy as np

@pytest.mark.skip(reason="Temporary Skip")
def testffinder():
    s = Smurfs(file='testFile.dat')
    s.run(4,f_min=15,skip_similar=True,mode='lmfit')
    #_ff.plot(show=True)
    s.plot_pdg(show=True)
    print(s._result)

@pytest.mark.skip(reason="Temporary Skip")
def testCreatedSin():
    x = np.linspace(0, 27, num=15000)
    params = [0.6,0.2,0,0.1,6,0,10,5,6]
    y = sin_multiple(x,*params) + np.random.normal(0,0.3,len(x))
    s = Smurfs(time=x, flux=y, label='test_sin')
    s.run(4,extend_frequencies=5,mode='lmfit')
    pass

def testGammaDoradus():
    s = Smurfs(target_name='TIC 3243257')
    s.run(snr=4,window_size=2,extend_frequencies=3)
    #_ff.plot(show=True)
    s.plot_pdg(show=True)
    print(s._result)



import sys

if not sys.warnoptions:
    import warnings

    warnings.simplefilter("ignore")

import lightkurve as lk
import matplotlib.pyplot as pl
import numpy as np
from pandas import DataFrame as df
import os
import pickle
on_rtd = os.environ.get('READTHEDOCS') == 'True'
if not on_rtd:
    from pyfcomb import get_combinations
from uncertainties import unumpy as unp
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from io import StringIO
from pandas import read_csv

from smurfs.preprocess.tess import download_lc
from smurfs.preprocess.file import load_file
from smurfs._smurfs.frequency_finder import FFinder, sin_multiple
from smurfs.signal.periodogram import Periodogram
from smurfs.signal.lightcurve import LightCurve
from smurfs.support.support import cd
import smurfs.support.mprint as mpr
from smurfs.support.mprint import *
from uncertainties import ufloat_fromstr
import matplotlib.backends.backend_pdf


class Smurfs:
    """
    The *Smurfs* class is the main way to start your frequency analysis. The workflow for a generic problem is the
    following:

    1) Instantiate the *Smurfs* class, by providing the light curve through one of different methods.
    2) Call 'run', by providing at least the signal to noise ratio and window size of the analysis.
    3) Either save the analysis using 'save' or get the result from the class and continue your analysis.

    After 'run' has finished, you can access the results through various channels:

    - Use the 'ff' property (returns the 'FFinder' instance, where the analysis happens)
    - Use the 'result' property

    The class also has other interesting properties like 'combinations' (calculates all possible combinations
    for the frequencies from the results, uses [pyfcomb](https://github.com/MarcoMuellner/pyfcomb) ), 'nyquist' (
    the nyquist frequency of the provided data) and more.

    You can also plot the results using the 'plot_lc' or 'plot_pdg' methods.

    You can provide a light curve through three different methods:

    1) Set *target_name*: Can be any star that has been observed by the TESS or Kepler mission. You can provide any TIC or KIC ID (including KIC/TIC) or any name resolvable by Simbad.
    2) Set *time* and *flux*
    3) Set *file*: Needs to be an ASCII file containing time and flux

    Either can be used. 1) and 2) will be automatically sigma clipped and converted to magnitude.

    :param file: ASCII file containing time and flux
    :param time: time column of the light curve
    :param flux: flux column of the light curve
    :param target_name: Name of the target, resolvable by Simbad or either KIC/TIC ID
    :param flux_type: If you supply a target name that has been observed by TESS SC mode, you can choose either 'PCDSAP' or 'SAP' flux for that target.
    :param label: Optional label for the star. Results will be saved under this name
    :param quiet_flag: Quiets Smurfs (no more print message will be piped to stdout)
    """
    def __init__(self, file=None, time=None, flux=None, target_name=None, flux_type='PDCSAP', label=None,
                 quiet_flag=False,mission = 'TESS',sigma_clip : float=4,iters :int=1,do_pca : bool=False,do_psf :bool = False, apply_file_correction = False):
        Settings.quiet = quiet_flag
        self.validation_page : Figure= None
        if target_name is not None:
            self.lc, self.validation_page = download_lc(target_name, flux_type,mission,sigma_clip,iters,do_pca,do_psf)
            if label is None:
                self.label = target_name
            else:
                self.label = label            
        elif time is None and flux is None and file is None:
            raise AttributeError(
                ctext("You need to either pass a target path or time and flux of the lightcurve object", error))

        elif time is not None and flux is not None:
            mprint("Creating light curve object from time and flux input.", log)
            self.lc: LightCurve = LightCurve(lk.LightCurve(time=time, flux=flux))
            if label is None:
                self.label = 'LC'
            else:
                self.label = label

            if not 0.1 > np.mean(flux) > -0.1:
                mprint("Be aware that the mean of the flux is not 0! This might lead to unintended consequences!",warn)

        elif file is not None:
            self.lc: LightCurve = load_file(file,sigma_clip,iters,apply_file_correction)
            if label is None:
                self.label = os.path.basename(file).split(".")[0]
            else:
                self.label = label

        else:
            raise AttributeError(
                ctext("You need to either pass a target path or time and flux of the lightcurve object", error))

        self.pdg: Periodogram = Periodogram.from_lightcurve(self.lc)
        self._result = df([], columns=['f_obj', 'frequency', 'amp', 'phase', 'snr', 'res_noise', 'significant'])
        self._combinations = df([],
                                columns=["Name", "ID", "Frequency", "Amplitude", "Solution", "Residual", "Independent",
                                         "Other_Solutions"])
        self._ff: FFinder = None
        self._spectral_window = None

        # Original light curve to perform some processsing on that
        self.original_lc = self.lc.copy()

        # Target settings
        self.target_name = target_name
        self.flux_type = flux_type

        # Frequency settings
        self.snr = np.nan
        self.window_size = np.nan
        self.f_min = np.nan
        self.f_max = np.nan
        self.skip_similar = None
        self.similar_chanel = None
        self.extend_frequencies = np.nan
        self._notes = None

        mprint(f"Duty cycle for {self.label}: {'%.2f' % (self.duty_cycle * 100)}%", info)

    @property
    def combinations(self):
        """
        Gives a pandas dataframe of all possible combinations for the frequencies from *result*. It consists
        of the following columns in this order:

        - Name: Name of the frequency
        - ID: Frequency ID (order in which it was removed from the light curve)
        - Frequency: The frequency for which combinations where searched.
        - Amplitude: Amplitude of the frequency
        - Solution: Best solution for this frequency
        - Residual: Residual for the best solution
        - Independent: Flag if the frequency is independent according to the solver
        - Other_solutions: All other possible solutions for this frequency

        Will be populated after *run* was called.
        """
        return self._combinations

    @property
    def result(self):
        """
        Gives a pandas dataframe of the result from smurfs. It consists of the following columns in this order:

        - f_obj: *Frequency* object, that represents a given frequency
        - frequency
        - amp
        - phase
        - snr
        - res_noise: Residual noise
        - significant: Flag that shows if a frequency is significant or not
        """
        return self._result

    @property
    def ff(self):
        """
        Returns the *FrequencyFinder* object.
        """
        return self._ff

    @property
    def settings(self):
        """
        Returns a dataframe consising of the settings used in the analysis.
        """
        columns = ['Signal to Noise Ratio',
                   'Window size',
                   'Lower frequency bound',
                   'Upper frequency bound',
                   'Skip similar frequency regions',
                   'Chancel run after 10 similar frequencies',
                   'Ignore unsignificant frequencies number',
                   ]
        return df([[self.snr, self.window_size, self.f_min, self.f_max, self.skip_similar, self.similar_chanel
                       , self.extend_frequencies]], columns=columns)

    @property
    def statistics(self):
        """
        Returns a dataframe consisting of various statistics of the run.
        """
        columns = ['Duty cycle',
                   'Nyquist frequency',
                   'Total number of found frequencies']
        return df([[self.duty_cycle, self.nyquist, len(self._result)]], columns=columns)

    @property
    def obs_length(self):
        """
        Returns the length of the data set.
        """
        return self.lc.time[-1] - self.lc.time[0]

    @property
    def nyquist(self):
        """
        Returns the nyquist frequency
        """
        return 1 / (2 * np.median(np.diff(self.lc.time)))

    @property
    def duty_cycle(self):
        """
        Shows the duty cycle of the light curve
        """
        diff = np.diff(self.lc.time)
        t = np.median(diff)
        std = np.std(diff)
        mask = diff > (t + 3 * std)
        return (1 - np.sum(diff[mask]) / np.sum(diff))

    @property
    def periodogram(self):
        """
        Returns a *Periodogram* object of the light curve.
        """
        return self.pdg

    @property
    def spectral_window(self):
        """
        Computes the spectral window of a given dataset by transforming the light curve with constant flux.
        """
        if self._spectral_window is None:
            spec_lc = self.lc.copy()
            spec_lc.flux = np.zeros(len(self.lc.flux)) + 1
            self._spectral_window = Periodogram.from_lightcurve(spec_lc)
        return self._spectral_window

    def fold(self, period, t0=None, transit_midpoint=None):
        """
        Returns a folded light curve. Signature equivalent to *lightkurve.LightCurve.fold*.
        """
        return self.lc.fold(period, t0, transit_midpoint)

    @property
    def notes(self):
        return self._notes

    @notes.setter
    def notes(self,value):
        self._notes = value

    def flatten(self, window_length=101, polyorder=2, return_trend=False, break_tolerance=5, niters=3, sigma=3,
                mask=None, **kwargs):
        """
        Flattens the light curve by applying a Savitzky Golay filter. Signature equivalent to
        *lightkurve.LightCurve.flatten*.
        """
        if return_trend:
            self.lc, self.trend = self.original_lc.flatten(window_length, polyorder, return_trend, break_tolerance,
                                                           niters,
                                                           sigma, mask, **kwargs)
        else:
            self.lc = self.original_lc.flatten(window_length, polyorder, return_trend, break_tolerance, niters, sigma,
                                               mask,
                                               **kwargs)

        self.lc = self.lc.remove_outliers(sigma,maxiters=niters)
        self.lc = self.lc.remove_nans()
        self.pdg: Periodogram = Periodogram.from_lightcurve(self.lc)

    def run(self, snr: float = 4, window_size: float = 2, f_min: float = None, f_max: float = None,
            skip_similar: bool = False, similar_chancel=True, extend_frequencies: int = 0, improve_fit=True,
            mode='lmfit'):
        """
        Starts the frequency analysis by instantiating a *FrequencyFinder* object and running it. After finishing the
        run, combinations are computed. See *FrequencyFinder.run* for an explanation of the algorithm.

        :param snr: Signal to noise ratio, that provides a lower end of the analysis.
        :param window_size: Window size, with which the SNR is computed.
        :param f_min: Minimum frequency that is considered in the analysis.
        :param f_max: Maximum frequency that is considered in the analysis.
        :param skip_similar: Flag that skips a certain range if too many similar frequencies in this range are found in a row.
        :param similar_chancel: Flat that chancels the run after 10 frequencies found that are too similar.
        :param extend_frequencies: Extends the analysis by this number of insignificant frequencies.
        :param improve_fit: If this flag is set, all combined frequencies are re-fitted after every new frequency was found
        :param mode: Fitting mode. You can choose between 'scipy' and 'lmfit'
        """

        self.snr = snr
        self.window_size = window_size
        self.f_min = f_min
        self.f_max = f_max
        self.skip_similar = skip_similar
        self.similar_chanel = similar_chancel
        self.extend_frequencies = 0

        self._ff = FFinder(self, f_min, f_max)
        self._result = self._ff.run(snr=snr, window_size=window_size, skip_similar=skip_similar,
                                    similar_chancel=similar_chancel
                                    , extend_frequencies=extend_frequencies, improve_fit=improve_fit, mode=mode)
        self._combinations = get_combinations((self._result[self._result.significant == True].index+1).tolist(),
                                              unp.nominal_values(
                                                  self._result[self._result.significant == True].frequency.tolist())
                                              , unp.nominal_values(
                self._result[self._result.significant == True].amp.tolist()))

        self.res_lc = self._ff.res_lc

        mprint(f"{self.label} Analysis done!",info)

    def improve_result(self):
        """
        Fits the combined found frequencies to the original light curve, hence improving the fit of the total model.
        """
        if self._ff is None:
            raise AttributeError("You need to run the analysis before you can improve the fit.")

        self._result = self._ff.improve_result()
        self._combinations = get_combinations((self._result[self._result.significant == True].index + 1).tolist(),
                                              unp.nominal_values(
                                                  self._result[self._result.significant == True].frequency.tolist())
                                              , unp.nominal_values(
                self._result[self._result.significant == True].amp.tolist()))
        self.res_lc = self._ff.res_lc


    def plot_lc(self, show=False, result = None, **kwargs):
        """
        Plots the light curve. If a result is already computed, it also plots the resulting model

        :param show: if this is set, pyplot.show() is called
        :param kwargs: kwargs for *lightkurve.LightCurve.scatter*
        """
        for i in ['color', 'ylabel', 'normalize']:
            try:
                del kwargs[i]
            except KeyError:
                pass

        ax: Axes = self.lc.scatter()

        model = result if result is not None and len(result) >0 else self.result

        if len(model) > 0:
            params = []

            for i, j, k in zip(model[self._result.significant==True].amp
                    , model[self._result.significant==True].frequency
                    , model[self._result.significant==True].phase):
                params.append(i.nominal_value)
                params.append(j.nominal_value)
                params.append(k.nominal_value)

            y = sin_multiple(self.lc.time, *params)
            ax.plot(self.lc.time, y, color='red', linewidth=1)

        if show:
            pl.show()

    def plot_pdg(self, show=False, plot_insignificant=False, **kwargs):
        """
        Plots the periodogram. If the result is already computed, it will mark the found frequencies in the
        periodogram.

        :param show: if this is set, pyplot.show() is called
        :param plot_insignificant: Flag, if set, insignififcant frequencies are marked in the perioodogram
        :param kwargs: kwargs for *lightkurve.Periodogram.plot*
        :return:
        """
        if self._ff is None:
            self.pdg.plot(color='k', ylabel='Amplitude [mag]', **kwargs)
        else:
            self._ff.plot(show=show, plot_insignificant=plot_insignificant, **kwargs)

    def save(self, path: str, store_obj=False):
        """
        Saves the result of the analysis to a given folder.

        :param path: Path where the result is stored
        :param store_obj: If this is set, the Smurfs object is stored, and can be later reloaded.
        """
        if not os.path.exists(path):
            raise IOError(ctext(f"'{path}' does not exist!", error))

        mprint("Saving results, this may take a bit ...", log)
        proj_path = os.path.join(path, self.label.replace(" ","_"))
        index = 1
        while True:
            if not os.path.exists(proj_path):
                break

            proj_path = os.path.join(path, self.label.replace(" ","_") + f"_{index}")
            index += 1

        os.makedirs(proj_path)
        with cd(proj_path):
            os.makedirs("data")
            os.makedirs("plots")

            with cd("data"):
                if self._result is not None:
                    frame: df = self._result.drop(columns=['f_obj'])
                    frame.index.name = 'f_nr'
                    df_list = [(self.settings, '#Settings'),
                               (self.statistics, '#Statistics'),
                               (frame, '#Result')]

                    with open('result.csv', 'w') as f:
                        for fr, comment in df_list:
                            f.write(f"{comment}\n")
                            fr.to_csv(f)
                            f.write("\n\n")

                    self._combinations.to_csv('combinations.csv')
                self.lc.to_csv("LC.txt",**{'index':False})
                self.pdg.to_csv("PS.txt")
                if self._ff is not None:
                    self._ff.res_lc.to_csv("LC_residual.txt",**{'index':False})
                    self._ff.res_pdg.to_csv("PS_residual.txt")

                if self._notes is not None:
                    with open("notes.txt",'w') as f:
                        f.writelines(self._notes)

                if store_obj:
                    pickle.dump(self, open("obj.smurfs", "wb"))

            with cd("plots"):
                images = [(self.lc, "LC.pdf"),
                          (self.pdg, "PS.pdf"),
                          ]

                if self._ff is not None:
                    images += [(self._ff.res_lc, "LC_residual.pdf"),
                               (self._ff.res_pdg, "PS_residual.pdf"),
                               (self._ff, "PS_result.pdf")
                               ]

                for obj, name in images:
                    fig, ax = pl.subplots(figsize=(16, 10))
                    if isinstance(obj, LightCurve):
                        obj.scatter(ax=ax)
                    else:
                        obj.plot(ax=ax, markersize=2)
                    pl.tight_layout()
                    fig.savefig(name)
                    pl.close()

                if self.validation_page is not None:
                    pdf = matplotlib.backends.backend_pdf.PdfPages("Validation_page.pdf")
                    for fig in self.validation_page: ## will open an empty extra figure :(
                        pdf.savefig(fig)
                    pdf.close()

                    for fig in self.validation_page:
                        pl.close(fig)

        mprint(f"{self.label} Data saved!",info)

    @staticmethod
    def from_path(path: str):
        """
        Loads a smurfs object from path. You need to have set the *store_obj* flag in *save*, for this object to be
        saved.
        """
        if not os.path.exists(path):
            raise IOError(ctext(f"'{path}' does not exist!", error))

        load_file = None

        for r, d, f, in os.walk(path):
            for file in f:
                if '.smurfs' in file:
                    load_file = os.path.join(r, file)

        if load_file is None:
            raise IOError(ctext(f"Can't find any .smurfs file in {path}!", error))

        return pickle.load(open(load_file, 'rb'))

    @staticmethod
    def load_results(path: str):
        """
        Loads the pandas dataframes from the results file
        :param path: exact path of the results file
        :return: 3 pandas dataframes, settings, statistics and results
        """

        if not os.path.exists(path) or not path.endswith("result.csv"):
            raise IOError("You need to provide the exact path to the results file")

        with open(path,'r') as f:
            content = f.read()

        settings = read_csv(StringIO(content.split("\n\n")[0]),skiprows=1)
        statistics = read_csv(StringIO(content.split("\n\n")[1]),skiprows=2)
        results = read_csv(StringIO(content.split("\n\n")[2]),skiprows=2,
                           converters = {'frequency':ufloat_fromstr,
                                         'amp':ufloat_fromstr,
                                         'phase':ufloat_fromstr})

        return settings,statistics,results
